package net.sf.jsequnit.javadoc;

import java.io.IOException;
import java.util.List;
import java.util.Map;

import net.sf.jsequnit.config.Configuration;
import net.sf.jsequnit.diagram.DiagramCreator;

import com.sun.javadoc.MethodDoc;
import com.sun.javadoc.ProgramElementDoc;
import com.sun.javadoc.Tag;
import com.sun.tools.doclets.Taglet;

/**
 * A Taglet processing tag {@code sequence.diagram}.
 * This tag is used to include sequence diagrams into the javadoc.
 * The sequence diagram is created through the utilization of JUnit test cases.
 * 
 * @author jacek.ratzinger
 *
 */
public class SequenceTaglet implements Taglet {
	
	/** The name of the tag. */
	public static final String TAG_NAME = "sequence.diagram";
	/** Parameter for the test method to be executed for tracing. */
	public static final String TEST_PARAMETER = "test";
	/** Parameter for the method that is the root of the generated sequence diagram. */
	public static final String TRACE_PARAMETER = "trace";
	/** Parameter to exclude the given packages, classes and methods. */
	public static final String EXCLUDE_PARAMETER = "omit";
	/** Parameter to indicate methods where sub calls should not be traced. */
	public static final String UPTO_PARAMETER = "upto";
	/** Parameter to indicate that only public methods should be included in the resulting sequence diagram. */
	public static final String PUBLIC_PARAMETER = "public-only";

	private static final String IMAGE_DIR = "sequence-diagrams";

	/**
	 * To be loaded and used by doclets at run-time,
	 * the taglet must have a static method called register
	 * that accepts a Map as an argument. 
	 * @see com.sun.tools.doclets.Taglet
	 * 
	 * @param tagletMap
	 *            the map to register this tag to.
	 */
	public static void register(Map tagletMap) {
       SequenceTaglet taglet = new SequenceTaglet();
       Taglet registeredTaglet = (Taglet) tagletMap.get(taglet.getName());
       if (registeredTaglet != null) {
           tagletMap.remove(taglet.getName());
       }
       tagletMap.put(taglet.getName(), taglet);
    }
	
	/**
	 * Returns the name of this tag.
	 */
	public String getName() {
		return TAG_NAME;
	}

	/**
	 * Returns true to indicate that this tag can be used in constructor documentation.
	 */
	public boolean inConstructor() {
		return true;
	}

	/**
	 * Return false to indicate that this tag cannot be used in field documentation.
	 */
	public boolean inField() {
		return false;
	}

	/**
	 * Returns true to indicate that this tag can be used in method documentation.
	 */
	public boolean inMethod() {
		return true;
	}

	/**
	 * Returns true to indicate that this tag can be used in overview documentation.
	 */
	public boolean inOverview() {
		return true;
	}

	/**
	 * Returns true to indicate that this tag can be used in package documentation.
	 */
	public boolean inPackage() {
		return true;
	}

	/**
	 * Returns true to indicate that this tag can be used in type documentation (classes or interfaces).
	 */
	public boolean inType() {
		return true;
	}

	/**
	 * Returns false to indicate that this is not an inline tag.
	 */
	public boolean isInlineTag() {
		return false;
	}

	/**
	 * If several tags are available, add a sequence diagram for each tag.
	 */
	public String toString(Tag[] arg0) {
		StringBuffer result = new StringBuffer();
		for (int i = 0; i < arg0.length; i++) {
			Tag tag = arg0[i];
			result.append(toString(tag));
		}
		return result.toString();
	}
	
	/**
	 * Create a sequence diagram based on the information of the JavaDoc tag.
	 * 
	 * This is the main method of the {@code Taglet}.
	 * First the a test method is run in a debug VM
	 * and the trace is used to generate the sequence diagram.
	 * Then a html snippet is returned that links the image file into the documentation.
	 * 
	 * The following sequence diagram shows the basic interactions of this method.
	 * 
	 * @sequence.diagram public-only
	 *   test=net.sf.jsequnit.javadoc.SequenceTagletTest.shouldCreateDiagramFile
	 *   omit=org.mockito.*,org.apache.*,org.xml.sax.*,th.co.edge.jseq.EventThread,th.co.edge.jseq.EventThread$ThreadTrace,net.sf.jsequnit.process.ProcessController$*,net.sf.sdedit.*,th.co.edge.jseq.ActivationList,th.co.edge.jseq.Activation,net.sf.jsequnit.jseq.sdedit.SdeditTextDiagram,net.sf.jsequnit.javadoc.TagParser
	 *
	 * @param tag
	 *            The Tag representation of this custom tag.
	 */
	public String toString(Tag tag) {
		if (getTestMethod(tag) == null) {
			System.err.println("ignore sequence diagram tag for method: " +
								tag.holder().name() +
								" - no test method defined");
			return "";
		}
		try {
			createSequenceDiagram(tag, getOutputFile(tag));
		} catch (Exception e) {
			throw new RuntimeException("diagram creation failed: " + tag.holder().name(), e);
		}
		return htmlSnippet(tag);
	}

	/**
	 * Use the information of the tag to identify the name of the sequence diagram file.
	 * Return an absolute path so that the image file is located
	 * next to the JavaDoc documentation of the class.
	 * 
	 * @param tag
	 *            The Tag representation of the custom tag.
	 * 
	 * @return The absolute path to the image file containing the sequence diagram.
	 */
	public String getOutputFile(Tag tag) {
		String ouputDirectory = Configuration.getInstance().getOutputDirectory();
		return ouputDirectory + "/" + packagePath(tag) + "/" + imageLocation(tag);
	}

	/**
	 * The html snippet to link the sequence diagram into the JavaDoc documentation.
	 * 
	 * @param tag
	 *            The Tag representation of the custom tag.
	 * 
	 * @return a simple html snippet referencing the image file of the sequence diagram.
	 */
	public String htmlSnippet(Tag tag) {
		return "\n" + "<DT><B>Sequence Diagram:</B></DT>" 
				+ "<DD><img src=\"" + imageLocation(tag) + "\"/><DD>";
	}
	
	private String imageLocation(Tag tag) {
		return IMAGE_DIR + "/" + getOutputFileName(tag);
	}
	
	private String packagePath(Tag tag) {
		if (tag.holder() instanceof ProgramElementDoc) {
			ProgramElementDoc programElementDoc = (ProgramElementDoc) tag.holder();
			return programElementDoc.containingPackage().name().replaceAll("\\.", "/");
		} else {
			throw new IllegalArgumentException("unknown programming element:" + tag.holder().name());
		}
	}
	
	private String getOutputFileName(Tag tag) {
		String qualifiedTraceMethod = getTraceMethod(tag);
		String qualifiedTestMethod = getTestMethod(tag);
		String outputFileName = getOutputFileName(qualifiedTestMethod, qualifiedTraceMethod);
		return outputFileName;
	}

	public String getOutputFileName(String qualifiedTestMethod, String qualifiedTraceMethod) {
		String[] traceMethodElements = qualifiedTraceMethod.split("\\.");
		String[] testMethodElements = qualifiedTestMethod.split("\\.");
		String traceMethodName = traceMethodElements[traceMethodElements.length - 1];
		String traceClassName = traceMethodElements[traceMethodElements.length - 2];
		String testMethodName = testMethodElements[testMethodElements.length - 1];
		return traceClassName + "-" + traceMethodName + "_" + testMethodName + ".png";
	}

	/**
	 * Generate a sequence diagram based on the information of the tag representation
	 * and write to the provided file.
	 * 
	 * @param tag
	 *            The Tag representation of the custom tag.
	 * @param outputFileName
	 *            The absolute file path for the resulting image file.
	 * 
	 * @throws NoSuchMethodException
	 *            if the methods given in the tag cannot be found (e.g. due to typos
	 * @throws IOException
	 *            if an error occurs when the sequence diagram is writen to file
	 */
	public void createSequenceDiagram(Tag tag, String outputFileName)
	throws NoSuchMethodException, IOException {
		String testMethod = getTestMethod(tag);
		String traceMethod = getTraceMethod(tag);
		List<String> traceExcludes = getTraceExcludes(tag);
		List<String> boundaryMethods = getBoundaryMethods(tag);
		boolean isPublicOnly = isPublicOnly(tag);
		
		DiagramCreator diagramCreator = new DiagramCreator();
		diagramCreator.createSequenceDiagram(
							testMethod, 
							traceMethod, 
							traceExcludes, 
							boundaryMethods,
							isPublicOnly,
							outputFileName);
	}

	private String getTestMethod(Tag tag) {
		return getMethodNameForParameter(tag, TEST_PARAMETER);
	}
	
	private String getTraceMethod(Tag tag) {
		return getMethodNameForParameter(tag, TRACE_PARAMETER);
	}
	
	private List<String> getTraceExcludes(Tag tag) {
		TagParser tagParser = new TagParser(tag);
		return tagParser.getParamValues(EXCLUDE_PARAMETER);
	}
	
	private List<String> getBoundaryMethods(Tag tag) {
		TagParser tagParser = new TagParser(tag);
		return tagParser.getParamValues(UPTO_PARAMETER);
	}
	
	private boolean isPublicOnly(Tag tag) {
		TagParser tagParser = new TagParser(tag);
		return tagParser.paramExist(PUBLIC_PARAMETER);
	}

	private String getMethodNameForParameter(Tag tag, String paramName) {
		TagParser tagParser = new TagParser(tag);
		String paramValue = tagParser.getParamValue(paramName);
		if (paramValue == null && tag.holder().isMethod()) {
			MethodDoc methodDoc = (MethodDoc) tag.holder();
			paramValue = methodDoc.qualifiedName();
		}
		if (paramValue != null) {
			paramValue = paramValue.replace("#", ".");
			paramValue = paramValue.replace("(", "");
			paramValue = paramValue.replace(")", "");
		}
		return paramValue;
	}

}
