package net.sf.jsequnit.junit;

import java.util.List;

import org.junit.runner.JUnitCore;
import org.junit.runner.Request;
import org.junit.runner.Result;
import org.junit.runner.notification.Failure;

/**
 * A simple wrapper to run a single JUnit test method
 * with the help of JUnitCore.
 * 
 * @author jacek.ratzinger
 *
 */
public class SingleTestMethodJUnit {

	/**
	 * The main method accepting a single parameter
	 * (i.e. the fully qualified name of the JUnit test method).
	 * 
	 * @param args
	 *            The arguments containing the test method name.
	 */
	public static void main(String[] args) {
		SingleTestMethodJUnit instance = new SingleTestMethodJUnit();
		System.exit(instance.run(args));
	}

	private int run(String[] args) {
		String qualifiedTestMethod = args[0];
		qualifiedTestMethod = qualifiedTestMethod.replaceAll("#", ".");
		Result testResult = runTest(qualifiedTestMethod);
		if (testResult.wasSuccessful()) {
			return 0;
		} else {
			List<Failure> failures = testResult.getFailures();
			for (Failure failure : failures) {
				System.err.println("test " + qualifiedTestMethod + " failed: " + failure.getMessage());
				Throwable throwable = failure.getException();
				if (throwable != null) {
					throwable.printStackTrace();
				}
			}
			return 1;
		}
	}

	/**
	 * Run the provided test method using a JUnit Request passed to JUnitCore.
	 * 
	 * The following sequence diagram shows the interactions of this method.
	 * 
	 * @sequence.diagram
	 *   test=net.sf.jsequnit.junit.SingleTestMethodJUnitTest.shouldRunSingleTest
	 *   upto=org.junit.runner.Request.method,org.junit.runner.JUnitCore.run
	 * 
	 * @param qualifiedTestMethod
	 *            The fully qualified test method name.
	 * @return The {@code Result} of the test, which contains all information about the test run.
	 */
	public Result runTest(String qualifiedTestMethod) {
		Class<?> jUnitTestClass = getClass(qualifiedTestMethod);
		String testMethodName = getMethodName(qualifiedTestMethod);
		Request testRunRequest = Request.method(jUnitTestClass, testMethodName);
		JUnitCore jUnitCore = new JUnitCore();
		return jUnitCore.run(testRunRequest);
	}

	private Class<?> getClass(String qualifiedMethod) {
		int lastDot = qualifiedMethod.lastIndexOf(".");
		String qualifiedClass = qualifiedMethod.substring(0, lastDot);
		Class<?> result = null;
		try {
			result = Class.forName(qualifiedClass);
		} catch (ClassNotFoundException cnfe) {
			System.err.println("Could not find class: " + qualifiedClass);
		}
		return result;
	}

	private String getMethodName(String qualifiedMethod) {
		int lastDot = qualifiedMethod.lastIndexOf(".");
		return qualifiedMethod.substring(lastDot + 1);
	}

}